#include <Streaming.h>

// Mereni otacek
#define RPM_COUNT_MAX 4

// Klouzavy prumer
#define RANGE_AVG_MAX 10

// Automaticke mereni
#define MEASURE_LENGTH 25000 //ms
#define MEASURE_DELAY 6000 //ms
#define MEASURE_COUNT_MAX 10

// Vicestavova regulace
#define STATE_RPM_ZERO 5400  // RPM - klidovy stav
//#define STATE_DIR -1       // 1 pokud pro vyssi vzdalenost je vyssi RPM, jinak -1
#define STATE_RPM_EPS -50    // epsilon pro RPM
#define STATE_CMS_EPS 5      // epsilon pro vzdalenost

boolean auto_measure;
unsigned long measure_start = 0;
unsigned int measure_count = 0;

volatile boolean STOP = true;

volatile byte rpm_count;
volatile unsigned long rpm;
volatile unsigned long rpm_last_time;

volatile unsigned long last_rpm;
int isr_count = 0;

volatile unsigned long range;
unsigned long range_avg[RANGE_AVG_MAX];
unsigned int range_avg_index = 0;

// PWM output @ 25 kHz, only on pins 9 and 10.
// Output value should be between 0 and 320, inclusive.
// Site: http://arduino.stackexchange.com/questions/25609/need-help-to-set-pwm-frequency-to-25khz
void analogWrite25k(int pin, int value)
{
  switch (pin) {
  case 9:
    OCR1A = value;
    break;
  case 10:
    // OCR1B = value;
    break;
  default:
    // no other pin will work
    break;
  }
}

const int PIN_FAN_PWM = 9;
const int PIN_UZ_TRIG = 4;
const int PIN_UZ_ECHO = 5;

double Setpoint;  	// command
double Input;  		// plant output
double Output;          // controller output

boolean useFF;

double Setpoint_cm;
double Input_cm;
double Output_cm;

double Output_cm_bias;

boolean useState;

// Limits
const double UMAX = 5;
const double UMIN = 0;

// -- these parameters should be user-adjustable
double kp = 0.000183;    // proportional gain
double ki = 0.000183;    // integral gain
double kd = 0;  	 // derivative gain
double N = 100;   	 // filter coefficients
double Ts = 0.001;       // This must match actual sampling time PID

double kp_cm = 4.5;      // proportional gain
double ki_cm = 0.0005;   // integral gain
double kd_cm = 0;  	 // derivative gain
double N_cm = 0;   	 // filter coefficients
double Ts_cm = 0;        // This must match actual sampling time PID

#include <PID_v1.h>

PID myPID(&Input, &Output, &Setpoint, kp, ki, kd, DIRECT);
PID uzPID(&Input_cm, &Output_cm, &Setpoint_cm, kp_cm, ki_cm, kd_cm, REVERSE);

unsigned long lastTime = 0;

void toggleStop() {
  if (!STOP) {
    STOP = !STOP;
    analogWrite25k(PIN_FAN_PWM, 0);
    Serial << micros() << ':' << "stp" << '\n';
  } else {

    // PID
    myPID.FlushVariables();
    
    // UZ
    for (int i = 0; i < RANGE_AVG_MAX; i++) {
      range_avg[i] = 0;
    }
    range_avg_index = 0;
    
    // RPM
    rpm = 0;
    last_rpm = 0;
    rpm_count = 0;
    measure_start = millis();
    STOP = !STOP;
    Serial << micros() << ':' << "sta" << '\n';
    Serial << micros() << ":" << "set" << ':' << Setpoint << '\n';
    Serial << micros() << ":" << "scm" << ':' << Setpoint_cm << '\n';
  }
}

void setup() {
  // Spusteni seriove linky
  Serial.begin(9600);
  
  // Prirazeni funkce pro pocitani otacek
  attachInterrupt(0, rpm_fun, RISING);
  
  // Vychozi hodnoty (RPM)
  rpm = 0;
  last_rpm = 0;
  rpm_count = 0;
  rpm_last_time = 0;
  
  // PWM
  // Configure Timer 1 for PWM @ 25 kHz.
  TCCR1A = 0;           // undo the configuration done by...
  TCCR1B = 0;           // ...the Arduino core library
  TCNT1  = 0;           // reset timer
  TCCR1A = _BV(COM1A1)  // non-inverted PWM on ch. A
    | _BV(COM1B1)       // same on ch; B
      | _BV(WGM11);     // mode 10: ph. correct PWM, TOP = ICR1
  TCCR1B = _BV(WGM13)   // ditto
    | _BV(CS10);        // prescaler = 1
  ICR1   = 320;         // TOP = 320

  pinMode(PIN_FAN_PWM, OUTPUT);
  pinMode(PIN_UZ_TRIG, OUTPUT);
  pinMode(PIN_UZ_ECHO, INPUT);

  // PID
  myPID.SetOutputLimits(0, 5);
  myPID.SetSampleTime(1);
  myPID.SetMode(AUTOMATIC);
  
  uzPID.SetOutputLimits(-17600/2, 17600/2);
  uzPID.SetSampleTime(40);
  uzPID.SetMode(AUTOMATIC);

  Serial << "#Nastaveni PID:\n";
  Serial << "##Kp = " << _FLOAT(kp, 8) << '\n';
  Serial << "##Ki = " << _FLOAT(ki, 8) << '\n';
  Serial << "##Kd = " << _FLOAT(kd, 8) << '\n';

  Setpoint = 8800;
  Serial << "#Setpoint = " << Setpoint << '\n';

  Serial << "#Nastaveni PID UZ:\n";
  Serial << "##Kp = " << _FLOAT(kp_cm, 8) << '\n';
  Serial << "##Ki = " << _FLOAT(ki_cm, 8) << '\n';
  Serial << "##Kd = " << _FLOAT(kd_cm, 8) << '\n';

  Setpoint_cm = 70;
  Serial << "#Setpoint = " << Setpoint_cm << '\n';

  Output_cm_bias = 5440;
  Serial << "#Output bias = " << Output_cm_bias << '\n';

  useFF = true;
  Serial << "#FeedForward " << (useFF ? "on" : "off") << '\n';

  auto_measure = false;
  Serial << "#AutoMeasure " << (auto_measure ? "on" : "off") << '\n';

  if (auto_measure) {
    Serial << "##Delka = " << MEASURE_LENGTH << '\n';
    Serial << "##Prodleva = " << MEASURE_DELAY << '\n';
    Serial << "##Pocet = " << MEASURE_COUNT_MAX << '\n';
  }
  
  useState = true;
  Serial << "#StateController " << (useState ? "on" : "off") << '\n';
  
  if (useState) {
    Serial << "##Klidove RPM = " << STATE_RPM_ZERO << '\n';
    Serial << "##Epsilon RPM = " << STATE_RPM_EPS << '\n';
    Serial << "##Epsilon CMS = " << STATE_CMS_EPS << '\n';
  }

  // PID Interrupt
  TIMSK0 |= _BV(OCIE0A);

  // Vynulovani prumeru
  for (int i = 0; i < RANGE_AVG_MAX; i++) {
    range_avg[i] = 0;
  }

  analogWrite25k(PIN_FAN_PWM, 0);
  delay(5000);
}

void loop() {
  if (Serial.available() > 0) {
    char data = Serial.read();

    if (data == 's') {
      toggleStop();
    }
  }

  if (STOP)
    return;

  Serial << micros() << ':' << "rpm" << ':' << rpm << '\n';
  Serial << micros() << ':' << "pwm" << ':' << (useFF ? Output * 64 + (Setpoint * 5 / 17600) : Output * 64) << '\n';

  if (auto_measure && (millis() - measure_start) >= MEASURE_LENGTH) {
    toggleStop();
    if (++measure_count < MEASURE_COUNT_MAX) {
      delay(MEASURE_DELAY);
      toggleStop();
    }
  }

  unsigned long now = millis();
  unsigned long timeChange = (now - lastTime);
  if(timeChange>=40) {

  digitalWrite(PIN_UZ_TRIG, LOW);
  delayMicroseconds(2);
  digitalWrite(PIN_UZ_TRIG, HIGH);
  delayMicroseconds(5);
  digitalWrite(PIN_UZ_TRIG, LOW);

  long rangePWM = pulseIn(PIN_UZ_ECHO, HIGH);
  range = long((double(rangePWM) - 104.8945) / 52.3209);

  Serial << micros() << ':' << "cms" << ':' << range << '\n';

  range_avg[range_avg_index++] = range;

  if (range_avg_index >= RANGE_AVG_MAX)
    range_avg_index = 0;

  unsigned long cur_range = 0;

  for (int i = 0; i < RANGE_AVG_MAX; i++) {
    cur_range += range_avg[i];
  }

  cur_range /= RANGE_AVG_MAX;

  Input_cm = cur_range;

  Serial << micros() << ':' << "cmx" << ':' << cur_range << '\n';

  if (useState) {
    if (cur_range < Setpoint_cm-STATE_CMS_EPS)
      Setpoint = STATE_RPM_ZERO+STATE_RPM_EPS;
    else if (cur_range < Setpoint_cm+STATE_CMS_EPS)
      Setpoint = STATE_RPM_ZERO;
    else if (cur_range < Setpoint_cm+2*STATE_CMS_EPS)
      Setpoint = STATE_RPM_ZERO-STATE_RPM_EPS;
    else if (cur_range < Setpoint_cm+3*STATE_CMS_EPS)
      Setpoint = STATE_RPM_ZERO-2*STATE_RPM_EPS;
    else if (cur_range < Setpoint_cm+4*STATE_CMS_EPS)
      Setpoint = STATE_RPM_ZERO-3*STATE_RPM_EPS;
    else
      Setpoint = STATE_RPM_ZERO-4*STATE_RPM_EPS;
  } else {
    uzPID.Compute();

    Setpoint = constrain(Output_cm+Output_cm_bias, 0, 17600);
  }

  Serial << micros() << ":" << "set" << ':' << Setpoint << '\n';

  lastTime=now;
  }

}

// Funkce pro pocitani otacek
void rpm_fun()
{
  if (STOP)
    return;

  rpm_count++;
  
  if (rpm_count >= RPM_COUNT_MAX) {
    unsigned long rpm_time = micros();
    rpm = 30000000/(rpm_time - rpm_last_time)*4;
    rpm_last_time = rpm_time;
    rpm_count = 0;
  }
}

// Funkce pro provedeni regulace
SIGNAL(TIMER0_COMPA_vect) {
  if (STOP)
    return;

  Input = rpm;

  myPID.Compute();

  analogWrite25k(PIN_FAN_PWM, int(constrain(useFF ? Output * 64 + (Setpoint * 5 / 17600) : Output * 64, 0, 320)));

  last_rpm = rpm;
}
